import unittest
from typing import List

from ddt import unpack, data, ddt


from pangtreebuild.affinity_tree import builders as at_builders
from pangtreebuild.affinity_tree import parameters as at_params
from pangtreebuild.pangenome import graph
from pangtreebuild.pangenome.parameters import msa


def sid(x): return msa.SequenceID(x)


def nid(x): return graph.NodeID(x)


def b(x): return graph.Base(x)


@ddt
class AffinityTreeGenerationTests(unittest.TestCase):

    @data((at_params.P(0.5), graph.Compatibility(0.836660026534076)),
          (at_params.P(1), graph.Compatibility(0.7)),
          (at_params.P(4), graph.Compatibility(0.6561)))
    @unpack
    def test_1_p_parameter_influence(self,
                                     p: at_params.P,
                                     expected_cutoff: graph.Compatibility):
        nodes = [graph.Node(node_id=nid(0), base=b('T'), aligned_to=None),
                 graph.Node(node_id=nid(1), base=b('A'), aligned_to=None),
                 graph.Node(node_id=nid(2), base=b('G'), aligned_to=None),
                 graph.Node(node_id=nid(3), base=b('A'), aligned_to=None),
                 graph.Node(node_id=nid(4), base=b('C'), aligned_to=None),
                 graph.Node(node_id=nid(5), base=b('A'), aligned_to=None),
                 graph.Node(node_id=nid(6), base=b('C'), aligned_to=None),
                 graph.Node(node_id=nid(7), base=b('G'), aligned_to=None),
                 graph.Node(node_id=nid(8), base=b('T'), aligned_to=None),
                 graph.Node(node_id=nid(9), base=b('A'), aligned_to=None)
                 ]

        sequences = {
            msa.SequenceID('seq0'):
                graph.Sequence(msa.SequenceID('seq0'),
                               [graph.SeqPath([*map(nid, [10, 11, 12, 13, 14, 15, 16, 17, 18, 9])])],
                               graph.SequenceMetadata({})),
            msa.SequenceID('seq1'):
                graph.Sequence(msa.SequenceID('seq1'),
                               [graph.SeqPath([*map(nid, [10, 11, 12, 13, 14, 15, 16, 17, 8, 9])])],
                               graph.SequenceMetadata({})),
            msa.SequenceID('seq2'):
                graph.Sequence(msa.SequenceID('seq2'),
                               [graph.SeqPath([*map(nid, [10, 11, 12, 13, 14, 15, 16, 7, 8, 9])])],
                               graph.SequenceMetadata({})),
            msa.SequenceID('seq3'):
                graph.Sequence(msa.SequenceID('seq3'),
                               [graph.SeqPath([*map(nid, [10, 11, 12, 3, 4, 5, 6, 7, 8, 9])])],
                               graph.SequenceMetadata({})),
            msa.SequenceID('seq4'):
                graph.Sequence(msa.SequenceID('seq3'),
                               [graph.SeqPath([*map(nid, [10, 11, 2, 3, 4, 5, 6, 7, 8, 9])])],
                               graph.SequenceMetadata({}))
        }

        poagraph = graph.Poagraph(nodes, sequences)

        consensus_path = graph.SeqPath([*map(nid, [10, 11, 12, 13, 14, 15, 16, 17, 18, 19])])
        compatibilities = poagraph.get_compatibilities(poagraph.get_sequences_ids(),
                                                       consensus_path,
                                                       p)

        actual_cutoff = at_builders._find_node_cutoff([c for c in compatibilities.values()], []).cutoff
        self.assertAlmostEqual(expected_cutoff.value, actual_cutoff.value)

    @data(
        # single compatibility value
        (0.5, [graph.Compatibility(0.5)]),

        # two compatibilities values
        (0.7, [graph.Compatibility(0.5), graph.Compatibility(0.7)]),
        (1, [graph.Compatibility(1), graph.Compatibility(0.45)]),
        (0.9, [graph.Compatibility(0.9), graph.Compatibility(0.5)]),

        # repeated values
        (0.7, [*map(graph.Compatibility, [0.5, 0.7, 0.7])]),
        (0.9, [*map(graph.Compatibility, [0.9, 0.5, 0.5])]),
        (1, [*map(graph.Compatibility, [0.45, 1, 0.45, 0.45])]),

        # many unique compatibilities values
        (.8, [*map(graph.Compatibility, [.3, .4, .8])]),
        (0.91, [*map(graph.Compatibility, [0.31, 0.32, 0.91, 0.92, 0.93, 0.97])]),
        (0.91, [*map(graph.Compatibility, [0.29, 0.3, 0.33, 0.91, 0.92, 0.93, 0.97])]),
        (1, [*map(graph.Compatibility, [0.81, 0.75, 0.8, 0.81, 1])]),
        (0.9, [*map(graph.Compatibility, [0.5, 0.9, 0.99])]),
        (0.7, [*map(graph.Compatibility, [0.2, 0.85, 0.7, 0.8])]),
        (0.99, [*map(graph.Compatibility, [0.99, 0.9, 0.99])]),
        (0.99, [*map(graph.Compatibility, [0.99])]),

        # repeated distance between values
        (.4, [*map(graph.Compatibility, [.3, .4, .5])]),

        # all the same values
        (.1, [*map(graph.Compatibility, [.1, .1, .1])])
    )
    @unpack
    def test_2_find_cutoff_no_so_far_values(self,
                                            expected_cutoff: float,
                                            compatibilities: List[graph.Compatibility]):
        actual_cutoff = at_builders._find_node_cutoff(compatibilities, []).cutoff
        self.assertEqual(expected_cutoff, actual_cutoff.value)

    def test_3_find_cutoff_no_compatibilities(self):
        with self.assertRaises(ValueError) as err:
            _ = at_builders._find_node_cutoff([], []).cutoff
            self.assertEqual(str(err.exception), """Empty compatibilities list.
                                                    Cannot find cutoff.""")

    @data(
        # guard <= all compatibilities
        (0.2, [0.2, 0.7, 0.8, 0.85], [0.1, 0.01, 0]),
        (0.7, [0.7, 0.85, 0.7, 0.8], [0.1, 0.01, 0]),
        (0.8, [0.7, 0.7, 0.85, 0.8], [0.85, 0.91, 1.0]),

        # guard > all compatibilities
        (0.6, [0.3, 0.6, 0.61, 0.61], [0.99]),  # big distance to guard
        (0.9, [0.2, 0.97, 0.98, 0.9], [0.99]),  # small distance to guard

        # guard between compatibilities
        (0.5, [0.2, 0.57, 0.58, 0.5], [0.55]),  # take smaller than guard
        (0.58, [0.2, 0.27, 0.58, 0.2], [0.55]),  # take greater than guard
        (0.55, [0.2, 0.58, 0.27, 0.55], [0.55])  # take equal to guard
    )
    @unpack
    def test_4_find_cutoff_with_so_far_values(self,
                                              expected_cutoff,
                                              compatibilities,
                                              so_far_cutoffs):
        compatibilities = [graph.Compatibility(c) for c in compatibilities]
        so_far_cutoffs = [graph.Compatibility(c) for c in so_far_cutoffs]
        actual_cutoff = at_builders._find_node_cutoff(compatibilities, so_far_cutoffs).cutoff
        self.assertEqual(expected_cutoff, actual_cutoff.value)


if __name__ == '__main__':
    unittest.main()



"""Objects to store multiple sequence alignment files content: MAF, PO."""

import csv
from io import StringIO
from pathlib import Path
from typing import Optional, Any, Dict, List


class Maf:
    """MSA in MAF format. https://genome.ucsc.edu/FAQ/FAQformat.html#format5

    Args:
        file_name: MAF file name, no requirement to have full path.
        file_content: MAF content as stream.

    Raises:
        ValueError: If MAF is incorrect.

    Attributes:
        file_name (Path): File name.
        filecontent (str): MSA (MAF or PO) content content as stream."""

    def __init__(self, file_content: StringIO, file_name: Optional[Path]):
        self.filename = file_name
        self._raise_exception_if_incorrect(file_content)
        self.filecontent = file_content

    @staticmethod
    def _raise_exception_if_incorrect(filecontent):
        if not filecontent:
            raise ValueError("Incorrect MAF - empty file.")


class Po:
    """Multialignment in PO format.

    https://github.com/meoke/pangtree/blob/master/Documentation.md#po-file-format-specification

    Args:
        file_name: MAF file name, no requirement to have full path.
        file_content: MAF content as stream.

    Raises:
        ValueError: If MAF is incorrect.

    Attributes:
        file_name (Path): File name.
        file_content (str): MSA (MAF or PO) content content as stream."""

    def __init__(self, file_content: StringIO, file_name: Optional[Path]):
        self.filename = file_name
        self._raise_exception_if_incorrect(file_content)
        self.filecontent = file_content

    @staticmethod
    def _raise_exception_if_incorrect(filecontent):
        if not filecontent:
            raise ValueError("Incorrect PO - empty file.")


class SequenceID(object):
    """ID of genome seuquence being a part of poagraph.

    Args:
        sequence_id: value of the sequence ID as str
        skip_part_before_dot: TODO remove this hack so that
                            always part before dot is taken as sequence ID

    Attributes:
        value (str): sequence ID
    """

    value: str

    def __init__(self, sequence_id: str):
        if len(sequence_id) == 0:
            raise ValueError("Sequence ID cannot be empty.")
        if sequence_id == "eboVir3.KM034562v1":  # Ebola data hack
            self.value = "KM034562v1"
        else:
            self.value = sequence_id.split('.')[0]

    def __str__(self):
        return f"{self.value}"

    def __repr__(self):
        return self.__str__()

    def __eq__(self, other: 'SequenceID'):
        return other and self.value.__eq__(other.value)

    def __hash__(self):
        return hash(self.value)

    def __le__(self, other: 'SequenceID'):
        return self.value.__le__(other.value)

    def __lt__(self, other: 'SequenceID'):
        return self.value.__lt__(other.value)

    def __ge__(self, other: 'SequenceID'):
        return self.value.__ge__(other.value)

    def __gt__(self, other: 'SequenceID'):
        return self.value.__gt__(other.value)


class MetadataCSV:
    """Metadata provides additional information about sequences.
    See example metadata file: data\\Ebola\\ebola_metadata.csv

    Args:
        filecontent: Content of the CSV file with metadata.
        filename: Name of the file.

    Attributes:
        metadata: Dictionary .
        filename: Name of the file.
    """

    def __init__(self, filecontent: StringIO, filename: Optional[Path]):
        filecontent = filecontent.read()
        self.filename: Path = filename
        self.metadata: Dict[SequenceID, Dict[str, Any]]
        self.metadata = self.csv_to_dict(filecontent)

    @classmethod
    def csv_to_dict(cls, filecontent: str):
        rd = csv.DictReader(StringIO(filecontent), delimiter=',')
        cls._raise_exception_if_incorrect(rd)

        d = {}
        for row in rd:
            seqid = SequenceID(row['seqid'])
            if seqid in d:
                raise ValueError("""Repeated values in seqid column in metadata file. Make them unique.""")
            d[seqid] = dict(row)
            if None in d[seqid].keys():
                raise ValueError("""CSV metadata error. Different number of columns in line 0 than in header line.""")
            del d[seqid]['seqid']
        return d

    @staticmethod
    def _raise_exception_if_incorrect(rd: csv.DictReader) -> None:
        headers = rd.fieldnames
        if not headers:
            raise ValueError('Empty csv file.')

        if 'seqid' not in headers:
            raise ValueError('No \'seqid\' column in metadata csv.')

        if headers.count('seqid') > 1:
            raise ValueError("""Only one \'seqid\' column in metadata csv is allowed.""")

    def get_all_sequences_ids(self) -> List[SequenceID]:
        return [*self.metadata.keys()]

    def get_sequence_metadata(self, seq_id) -> Dict[str, Any]:
        if seq_id in self.metadata:
            return self.metadata[seq_id]
        return {}

    def get_metadata_keys(self) -> List[str]:
        if self.metadata.values():
            example_metadata = [*self.metadata.values()][0]
            if example_metadata:
                return [*example_metadata.keys()]
        return []



# -*- coding: utf-8 -*-
from Bio import AlignIO
from .graph.Block import Block
from .graph.SequenceInfo import SequenceInfo

def start_position(sequence):
    # Return start position relative to the plus strand
    if sequence.annotations["strand"] == -1:
        return sequence.annotations["srcSize"] - sequence.annotations["start"] - sequence.annotations["size"]
    else:
        return sequence.annotations["start"]
    
def read_maf(maf_file):
    blocks, seq = [], []
    for i, mafblock in enumerate(AlignIO.parse(maf_file, "maf")):
        blocks.append(Block(i, mafblock))
        for sequence in mafblock:
            seq.append(SequenceInfo(i, sequence.id, start_position(sequence), sequence.annotations["strand"]))
    seq.sort(key = lambda s: (s.id, s.start))
    return blocks, seq



# -*- coding: utf-8 -*-
class SequenceInfo:
    def __init__(self, block_id, seq_id, start_pos, strand):
        self.block = block_id
        self.id = seq_id
        self.start = start_pos
        self.strand = strand